# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import l3_const
import networking_cisco.plugins
from networking_cisco.plugins.cisco.db.l3 import l3_router_appliance_db
from networking_cisco.plugins.cisco.db.l3 import routertype_db
from networking_cisco.plugins.cisco.db.scheduler import (
    l3_routertype_aware_schedulers_db as router_sch_db)
from networking_cisco.plugins.cisco.extensions import routerhostingdevice
from networking_cisco.plugins.cisco.extensions import routerrole
from networking_cisco.plugins.cisco.extensions import routertype

from neutron.api.v2 import attributes
from neutron.common import test_lib
from neutron.db import common_db_mixin
from neutron.extensions import l3
from neutron.extensions import standardattrdescription


L3_PLUGIN_KLASS = (
    'networking_cisco.tests.unit.cisco.l3.l3_router_test_support.'
    'TestL3RouterServicePlugin')
extensions_path = networking_cisco.plugins.__path__[0] + '/cisco/extensions'


class L3RouterTestSupportMixin(object):

    _is_ha_tests = False

    def _mock_get_routertype_scheduler_always_none(self):
        self.get_routertype_scheduler_fcn_p = mock.patch(
            'networking_cisco.plugins.cisco.db.l3.l3_router_appliance_db.'
            'L3RouterApplianceDBMixin._get_router_type_scheduler',
            mock.Mock(return_value=None))
        self.get_routertype_scheduler_fcn_p.start()

    def _mock_backlog_processing(self, plugin):
        mock.patch.object(plugin, '_is_master_process',
                          return_value=True).start()
        mock.patch.object(plugin, '_setup_backlog_handling').start()

    def _add_router_plugin_ini_file(self):
        # includes config file for router service plugin
        if hasattr(self, '_is_ha_tests') and self._is_ha_tests is True:
            cfg_file = (
                networking_cisco.__path__[0] +
                '/tests/unit/cisco/etc/ha/cisco_router_plugin.ini')
        else:
            cfg_file = (
                networking_cisco.__path__[0] +
                '/tests/unit/cisco/etc/cisco_router_plugin.ini')
        if 'config_files' in test_lib.test_config:
            test_lib.test_config['config_files'].append(cfg_file)
        else:
            test_lib.test_config['config_files'] = [cfg_file]


class TestL3RouterBaseExtensionManager(object):

    def get_resources(self):
        # Add the resources to the global attribute map
        # This is done here as the setup process won't
        # initialize the main API router which extends
        # the global attribute map
        # first, add hosting device attribute to router resource
        l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
            routerhostingdevice.EXTENDED_ATTRIBUTES_2_0['routers'])
        # also add role attribute to router resource
        l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
            routerrole.EXTENDED_ATTRIBUTES_2_0['routers'])
        # also add routertype attribute to router resource
        l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
            routertype.EXTENDED_ATTRIBUTES_2_0['routers'])
        # also add description attribute to router and fip resources
        ext_res = (standardattrdescription.Standardattrdescription().
                   get_extended_resources("2.0"))
        if 'routers' in ext_res:
            l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
                    ext_res['routers'])
        if 'floatingips' in ext_res:
            (l3_const.RESOURCE_ATTRIBUTE_MAP['floatingips'].
             update(ext_res['floatingips']))
        # finally, extend the global attribute map
        attributes.RESOURCE_ATTRIBUTE_MAP.update(
            l3_const.RESOURCE_ATTRIBUTE_MAP)
        res = l3.L3.get_resources()
        # add routertype resource
        for item in routertype.Routertype.get_resources():
            res.append(item)
        return res

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


# A L3 routing service plugin class supporting the routertype and
# routerhost:hostingdevice extensions
class TestL3RouterServicePlugin(
    common_db_mixin.CommonDbMixin,
    routertype_db.RoutertypeDbMixin,
    l3_router_appliance_db.L3RouterApplianceDBMixin,
    # we need the router scheduling db but do not expose the scheduling
    # REST operations
        router_sch_db.L3RouterTypeAwareSchedulerDbMixin):

    supported_extension_aliases = [
        "router",
        "standard-attr-description",
        routerhostingdevice.ROUTERHOSTINGDEVICE_ALIAS,
        routerrole.ROUTERROLE_ALIAS,
        routertype.ROUTERTYPE_ALIAS]

    def get_plugin_type(self):
        return bc.constants.L3

    def get_plugin_description(self):
        return "L3 Routing Service Plugin for testing"

    def cleanup_after_test(self):
        """This function should be called in the TearDown() function of
        test classes that use the plugin.

        Reset all class variables to their default values.
        This is needed to avoid tests to pollute subsequent tests.
        """
        TestL3RouterServicePlugin._router_schedulers = {}
        TestL3RouterServicePlugin._router_drivers = {}
        TestL3RouterServicePlugin._namespace_router_type_id = None
        TestL3RouterServicePlugin._backlogged_routers = set()
        TestL3RouterServicePlugin._refresh_router_backlog = True



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import l3_const
from networking_cisco.plugins.cisco.device_manager.plugging_drivers import (
    hw_vlan_trunking_driver as hw_vlan)
from networking_cisco.plugins.cisco.extensions import routerrole

LOG = logging.getLogger(__name__)


APIC_OWNED = 'apic_owned_'
APIC_SNAT_SUBNET = 'host-snat-pool-for-internal-use'
APIC_SNAT_NET = 'host-snat-network-for-internal-use'
EXTERNAL_GW_INFO = l3_const.EXTERNAL_GW_INFO
UUID_REGEX = '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}'
DEVICE_OWNER_ROUTER_GW = bc.constants.DEVICE_OWNER_ROUTER_GW
DEVICE_OWNER_ROUTER_INTF = bc.constants.DEVICE_OWNER_ROUTER_INTF
ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR


ACI_ASR1K_DRIVER_OPTS = [
    cfg.StrOpt('aci_transit_nets_config_file', default=None,
               help=_("ACI with ASR transit network configuration file.")),
]

cfg.CONF.register_opts(ACI_ASR1K_DRIVER_OPTS, "general")

DEFAULT_EXT_DICT = {'gateway_ip': '1.103.2.254',
                    'cidr_exposed': '1.103.2.1/24'}


class AciDriverConfigInvalidFileFormat(n_exc.BadRequest):
    message = _("The ACI Driver config file format is invalid")


class AciDriverConfigMissingGatewayIp(n_exc.BadRequest):
    message = _("The ACI Driver config is missing a gateway_ip "
                "parameter for %(ext_net)s.")


class AciDriverConfigMissingCidrExposed(n_exc.BadRequest):
    message = _("The ACI Driver config is missing a cidr_exposed "
                "parameter for %(ext_net)s.")


class AciDriverConfigMissingSegmentationId(n_exc.BadRequest):
    message = _("The ACI Driver config is missing a segmentation_id "
                "parameter for %(ext_net)s.")


class AciDriverNoAciDriverInstalledOrConfigured(n_exc.BadRequest):
    message = _("The ACI plugin driver is either not installed or "
                "the neutron configuration is incorrect.")


class AciVLANTrunkingPlugDriver(hw_vlan.HwVLANTrunkingPlugDriver):
    """Driver class for Cisco ACI-based devices.

    The driver works with VLAN segmented Neutron networks. It
    determines which workflow is active (GBP or Neutron), and
    uses that implementation to get the information needed for
    the networks between the hosting device and the ACI fabric.
    """
    # once initialized _device_network_interface_map is dictionary
    _device_network_interface_map = None
    _apic_driver = None
    _l3_plugin = None

    def __init__(self):
        super(AciVLANTrunkingPlugDriver, self).__init__()
        self._cfg_file = cfg.CONF.general.aci_transit_nets_config_file
        self._get_ext_net_name = None
        self._default_ext_dict = DEFAULT_EXT_DICT
        self._transit_nets_cfg = {}
        self._get_vrf_context = None

    def _sanity_check_config(self, config):
        for network in config.keys():
            if config.get(network).get('gateway_ip') is None:
                raise AciDriverConfigMissingGatewayIp(ext_net=network)
            if config.get(network).get('cidr_exposed') is None:
                raise AciDriverConfigMissingCidrExposed(ext_net=network)

    @property
    def transit_nets_cfg(self):
        if self._cfg_file:
            networks_dict = open(self._cfg_file, 'r').read()
            try:
                self._transit_nets_cfg = eval(networks_dict)
                self._sanity_check_config(self._transit_nets_cfg)
            except SyntaxError:
                raise AciDriverConfigInvalidFileFormat
        else:
            self._transit_nets_cfg = {}
        return self._transit_nets_cfg

    @property
    def get_ext_net_name(self):
        # Ensure that we have an APIC driver
        if self.apic_driver:
            return self._get_ext_net_name

    @property
    def get_vrf_context(self):
        if self.apic_driver:
            return self._get_vrf_context

    def _get_vrf_context_gbp(self, context, router_id, port_db):
        l3p_id = self.apic_driver.gbp_plugin.get_l3p_id_from_router_id(
            context.elevated(), router_id)
        if l3p_id:
            return {'vrf_id': l3p_id,
                    'vrf_name': l3p_id,
                    'vrf_tenant': None}

    def _get_vrf_context_neutron(self, context, router_id, port_db):
        router = self.l3_plugin.get_router(context, router_id)
        vrf_info = self.apic_driver.get_router_vrf_and_tenant(router)
        details = {}
        if self.apic_driver.per_tenant_context:
            details['vrf_id'] = router['tenant_id']
        else:
            details['vrf_id'] = str(vrf_info['aci_name'])
        details['vrf_name'] = vrf_info['aci_name']
        details['vrf_tenant'] = vrf_info['aci_tenant']
        return details

    def _get_external_network_dict(self, context, port_db):
        """Get external network information

        Get the information about the external network,
        so that it can be used to create the hidden port,
        subnet, and network.
        """
        if port_db.device_owner == DEVICE_OWNER_ROUTER_GW:
            network = self._core_plugin.get_network(context,
                port_db.network_id)
        else:
            router = self.l3_plugin.get_router(context,
                port_db.device_id)
            ext_gw_info = router.get(EXTERNAL_GW_INFO)
            if not ext_gw_info:
                return {}, None
            network = self._core_plugin.get_network(context,
                ext_gw_info['network_id'])

        # network names in GBP workflow need to be reduced, since
        # the network may contain UUIDs
        external_network = self.get_ext_net_name(network['name'])
        # TODO(tbachman): see if we can get rid of the default
        transit_net = self.transit_nets_cfg.get(
            external_network) or self._default_ext_dict
        transit_net['network_name'] = external_network
        return transit_net, network

    @property
    def l3_plugin(self):
        if not self._l3_plugin:
            self._l3_plugin = bc.get_plugin(bc.constants.L3)
        return self._l3_plugin

    @property
    def apic_driver(self):
        """Get APIC driver

        There are different drivers for the GBP workflow
        and Neutron workflow for APIC. First see if the GBP
        workflow is active, and if so get the APIC driver for it.
        If the GBP service isn't installed, try to get the driver
        from the Neutron (APIC ML2) workflow.
        """
        if not self._apic_driver:
            try:
                self._apic_driver = (bc.get_plugin(
                    'GROUP_POLICY').policy_driver_manager.
                                     policy_drivers['apic'].obj)
                self._get_ext_net_name = self._get_ext_net_name_gbp
                self._get_vrf_context = self._get_vrf_context_gbp
            except AttributeError:
                    LOG.info("GBP service plugin not present -- will "
                             "try APIC ML2 plugin.")
            if not self._apic_driver:
                try:
                    self._apic_driver = (
                        self._core_plugin.mechanism_manager.mech_drivers[
                            'cisco_apic_ml2'].obj)
                    self._get_ext_net_name = self._get_ext_net_name_neutron
                    self._get_vrf_context = self._get_vrf_context_neutron
                except KeyError:
                    LOG.error("APIC ML2 plugin not present: "
                              "no APIC ML2 driver could be found.")
                    raise AciDriverNoAciDriverInstalledOrConfigured()
        return self._apic_driver

    def _snat_subnet_for_ext_net(self, context, subnet, net):
        """Determine if an SNAT subnet is for this external network.

        This method determines if a given SNAT subnet is intended for
        the passed external network.

        For APIC ML2/Neutron workflow, SNAT subnets are created on
        a separate network from the external network. The association
        with an external network is made by putting the name of the
        external network in the name of the SNAT network name, using
        a well-known prefix.
        """
        if subnet['network_id'] == net['id']:
            return True

        network = self._core_plugin.get_network(
            context.elevated(), subnet['network_id'])
        ext_net_name = network['name']
        if (APIC_SNAT_NET + '-') in ext_net_name:
            # This is APIC ML2 mode -- we need to strip the prefix
            ext_net_name = ext_net_name[len(APIC_SNAT_NET + '-'):]
            if net['id'] == ext_net_name:
                return True
        return False

    def _add_snat_info(self, context, router, net, hosting_info):
        if net:
            snat_ips = self.apic_driver.get_snat_ip_for_vrf(context,
                router['tenant_id'], net)
            snat_subnets = self._core_plugin.get_subnets(
                context.elevated(), {'name': [APIC_SNAT_SUBNET]})
            if snat_subnets and snat_ips:
                hosting_info['snat_subnets'] = []
                for subnet in snat_subnets:
                    # Skip any SNAT subnets that aren't on this external
                    # network
                    if not self._snat_subnet_for_ext_net(context, subnet, net):
                        continue
                    snat_subnet = {'id': router['tenant_id'],
                                   'ip': snat_ips['host_snat_ip'],
                                   'cidr': subnet['cidr']}
                    hosting_info['snat_subnets'].append(snat_subnet)

    def extend_hosting_port_info(self, context, port_db, hosting_device,
                                 hosting_info):
        """Get the segmenetation ID and interface

        This extends the hosting info attribute with the segmentation ID
        and physical interface used on the external router to connect to
        the ACI fabric. The segmentation ID should have been set already
        by the call to allocate_hosting_port, but if it's not present, use
        the value from the port resource.
        """
        if hosting_info.get('segmentation_id') is None:
            LOG.debug('No segmentation ID in hosting_info -- assigning')
            hosting_info['segmentation_id'] = (
                port_db.hosting_info.get('segmentation_id'))
        is_external = (port_db.device_owner == DEVICE_OWNER_ROUTER_GW)
        hosting_info['physical_interface'] = self._get_interface_info(
            hosting_device['id'], port_db.network_id, is_external)
        ext_dict, net = self._get_external_network_dict(context, port_db)
        if is_external and ext_dict:
            hosting_info['network_name'] = ext_dict['network_name']
            hosting_info['cidr_exposed'] = ext_dict['cidr_exposed']
            hosting_info['gateway_ip'] = ext_dict['gateway_ip']
            details = self.get_vrf_context(context,
                                           port_db['device_id'], port_db)
            router_id = port_db.device_id
            router = self.l3_plugin.get_router(context, router_id)
            # skip routers not created by the user -- they will have
            # empty-string tenant IDs
            if router.get(ROUTER_ROLE_ATTR):
                return
            hosting_info['vrf_id'] = details['vrf_id']
            if ext_dict.get('global_config'):
                hosting_info['global_config'] = (
                    ext_dict['global_config'])
            self._add_snat_info(context, router, net, hosting_info)
        else:
            if ext_dict.get('interface_config'):
                hosting_info['interface_config'] = ext_dict['interface_config']

    def allocate_hosting_port(self, context, router_id, port_db, network_type,
                              hosting_device_id):
        """Get the VLAN and port for this hosting device

        The VLAN used between the APIC and the external router is stored
        by the APIC driver.  This calls into the APIC driver to first get
        the ACI VRF information associated with this port, then uses that
        to look up the VLAN to use for this port to the external router
        (kept as part of the L3 Out policy in ACI).
        """
        # If this is a router interface, the VLAN comes from APIC.
        # If it's the gateway, the VLAN comes from the segment ID
        if port_db.get('device_owner') == DEVICE_OWNER_ROUTER_GW:
            ext_dict, net = self._get_external_network_dict(context, port_db)
            # If an OpFlex network is used on the external network,
            # the actual segment ID comes from the config file
            if net and net.get('provider:network_type') == 'opflex':
                if ext_dict.get('segmentation_id'):
                    return {'allocated_port_id': port_db.id,
                            'allocated_vlan': ext_dict['segmentation_id']}
                else:
                    raise AciDriverConfigMissingSegmentationId(ext_net=net)
            return super(AciVLANTrunkingPlugDriver,
                         self).allocate_hosting_port(
                             context, router_id,
                             port_db, network_type, hosting_device_id)

        # shouldn't happen, but just in case
        if port_db.get('device_owner') != DEVICE_OWNER_ROUTER_INTF:
            return

        # get the external network that this port connects to.
        # if there isn't an external gateway yet on the router,
        # then don't allocate a port

        router = self.l3_plugin.get_router(context, router_id)
        gw_info = router[EXTERNAL_GW_INFO]
        if not gw_info:
            return
        network_id = gw_info.get('network_id')

        networks = self._core_plugin.get_networks(
            context.elevated(), {'id': [network_id]})
        l3out_network = networks[0]
        l3out_name = self.get_ext_net_name(l3out_network['name'])
        # For VLAN apic driver provides VLAN tag
        details = self.get_vrf_context(context, router_id, port_db)
        if details is None:
            LOG.debug('aci_vlan_trunking_driver: No vrf_details')
            return
        vrf_name = details.get('vrf_name')
        vrf_tenant = details.get('vrf_tenant')
        allocated_vlan = self.apic_driver.l3out_vlan_alloc.get_vlan_allocated(
            l3out_name, vrf_name, vrf_tenant=vrf_tenant)
        if allocated_vlan is None:
            if not vrf_tenant:
                # TODO(tbachman): I can't remember why this is here
                return super(AciVLANTrunkingPlugDriver,
                             self).allocate_hosting_port(
                                 context, router_id,
                                 port_db, network_type, hosting_device_id
                             )
            # Database must have been messed up if this happens ...
            return
        return {'allocated_port_id': port_db.id,
                'allocated_vlan': allocated_vlan}

    # TODO(tbahcman): get these from the drivers
    def _get_ext_net_name_gbp(self, network_name):
        """Get the external network name

        The name of the external network used in the APIC
        configuration file can be different from the name
        of the external network in Neutron, especially using
        the GBP workflow
        """
        prefix = network_name[:re.search(UUID_REGEX, network_name).start() - 1]
        return prefix.strip(APIC_OWNED)

    def _get_ext_net_name_neutron(self, network_name):
        """Get the external network name

        For Neutron workflow, the network name is returned
        as-is.
        """
        return network_name



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from networking_cisco.backwards_compatibility import ml2_config  # noqa
from networking_cisco.ml2_drivers.ucsm import (config as
    ucsm_config)
from networking_cisco.ml2_drivers.ucsm import constants as const
from networking_cisco.tests import base as nc_base

from neutron.common import config as neutron_config

from neutron.tests import base


UCSM_IP_ADDRESS_1 = '1.1.1.1'
UCSM_USERNAME_1 = 'username1'
UCSM_PASSWORD_1 = 'password1'
UCSM_VIRTIO_ETH_PORTS_1 = ['eth0, eth1']
UCSM_HOST_LIST_1 = ['UCS-1:UCS-1-SP, UCS-2:org-root/test/ls-UCS-2-SP']

UCSM_IP_ADDRESS_2 = '2.2.2.2'
UCSM_USERNAME_2 = 'username2'
UCSM_PASSWORD_2 = 'password2'
UCSM_VIRTIO_ETH_PORTS_2 = ['eth2, eth3']

UCSM_PHY_NETS = ['test_physnet']


ucsm_test_config_file = """
[ml2_cisco_ucsm]
supported_pci_devs=thing1:thing2, thing1:thing3
ucsm_https_verify=False

ucsm_ip=3.3.3.3
ucsm_username=username3
ucsm_password=password3
ucsm_host_list=UCS-3:UCS-3-SP

sriov_qos_policy=Global

[ml2_cisco_ucsm_ip:1.1.1.1]
ucsm_username=username1
ucsm_password=password1
ucsm_host_list=UCS-1:UCS-1-SP, UCS-2:org-root/test/ls-UCS-2-SP
ucsm_virtio_eth_ports=eth4, eth5
vnic_template_list=test-physnet:org-root:Test-VNIC,vnic2
sriov_qos_policy=Test

[ml2_cisco_ucsm_ip:2.2.2.2]
ucsm_username=username2
ucsm_password=password2
ucsm_virtio_eth_ports=eth2, eth3
vnic_template_list=physnet2:org-root/org-Test-Sub:Test
sp_template_list=SP_Template1_path:SP_Template1:S1,S2 \
                 SP_Template2_path:SP_Template2:S3,S4
sriov_qos_policy=

[ml2_cisco_ucsm_ip:4.4.4.4]
ucsm_username=username4
ucsm_password=password4

[sriov_multivlan_trunk]
test_network1=5, 7-9
test_network2=500-509, 700
"""

ucsm_config_bad_pci_devs = """
[ml2_cisco_ucsm]
supported_pci_devs=thing:thing, thing2, thing3:thing4
"""

CONF = cfg.CONF


class UCSMConfigTestCase(nc_base.TestCase):

    def setUp(self):
        super(UCSMConfigTestCase, self).setUp()
        CONF.reset()
        nc_base.load_config_file(ucsm_test_config_file)
        ucsm_config.load_single_ucsm_config()

    def test_oslo_config_configuration_loading(self):
        expected_config_data = {
            "ucsm_https_verify": False,
            "ucsm_ip": "3.3.3.3",
            "ucsm_username": "username3",
            "ucsm_password": "password3",
            "supported_pci_devs": ["thing1:thing2", "thing1:thing3"],
            "ucsm_host_list": {"UCS-3": "UCS-3-SP"},
            "ucsm_virtio_eth_ports": [const.ETH_PREFIX + const.ETH0,
                                      const.ETH_PREFIX + const.ETH1],
            "sriov_qos_policy": 'Global',
            "vnic_template_list": {},
            "sp_template_list": {},
            "ucsms": {
                "1.1.1.1": {
                    "ucsm_username": "username1",
                    "ucsm_password": "password1",
                    "ucsm_virtio_eth_ports": [const.ETH_PREFIX + "eth4",
                                              const.ETH_PREFIX + "eth5"],
                    "ucsm_host_list": {"UCS-1": "UCS-1-SP",
                                       "UCS-2": "org-root/test/ls-UCS-2-SP"},
                    "sriov_qos_policy": "Test",
                    "vnic_template_list": {
                        "test-physnet": ucsm_config.UCSTemplate(
                            "org-root", "Test-VNIC,vnic2")},
                    "sp_template_list": {},
                },
                "2.2.2.2": {
                    "ucsm_username": "username2",
                    "ucsm_password": "password2",
                    "ucsm_virtio_eth_ports": [const.ETH_PREFIX + "eth2",
                                              const.ETH_PREFIX + "eth3"],
                    "ucsm_host_list": None,
                    "sriov_qos_policy": '',
                    "vnic_template_list": {
                        "physnet2": ucsm_config.UCSTemplate(
                            "org-root/org-Test-Sub", "Test")},
                    "sp_template_list": {
                        "S1": ucsm_config.UCSTemplate("SP_Template1_path",
                                                      "SP_Template1"),
                        "S2": ucsm_config.UCSTemplate("SP_Template1_path",
                                                      "SP_Template1"),
                        "S3": ucsm_config.UCSTemplate("SP_Template2_path",
                                                      "SP_Template2"),
                        "S4": ucsm_config.UCSTemplate("SP_Template2_path",
                                                      "SP_Template2"),
                    },
                },
                "3.3.3.3": {
                    "ucsm_username": "username3",
                    "ucsm_password": "password3",
                    "ucsm_virtio_eth_ports": [const.ETH_PREFIX + "eth0",
                                              const.ETH_PREFIX + "eth1"],
                    "ucsm_host_list": {"UCS-3": "UCS-3-SP"},
                    "sriov_qos_policy": 'Global',
                    "vnic_template_list": {},
                    "sp_template_list": {},
                },
                # 4.4.4.4 Test's if we've inherited sriov_qos_policy from the
                # main group
                "4.4.4.4": {
                    "ucsm_username": "username4",
                    "ucsm_password": "password4",
                    "ucsm_virtio_eth_ports": [const.ETH_PREFIX + "eth0",
                                              const.ETH_PREFIX + "eth1"],
                    "ucsm_host_list": None,
                    "sriov_qos_policy": 'Global',
                    "vnic_template_list": {},
                    "sp_template_list": {},
                }
            }
        }

        expected_sriov_multivlan = {
            "test_network1": [5, 7, 8, 9],
            "test_network2": [500, 501, 502, 503, 504, 505, 506, 507,
                              508, 509, 700]
        }

        # Convert nested oslo.config GroupAttrs to dict for comparision
        loaded_config = dict(CONF.ml2_cisco_ucsm)
        for ip, data in CONF.ml2_cisco_ucsm.ucsms.items():
            loaded_config['ucsms'][ip] = dict(data)

        # Test the ml2_cisco_ucsm group is what we expect
        self.assertEqual(expected_config_data, loaded_config)

        # Test the sriov multivlan group is what we expect
        self.assertEqual(expected_sriov_multivlan,
                         dict(CONF.sriov_multivlan_trunk.network_vlans))

    def _assert_sp_templates_in_start_state(self):
        self.assertEqual(
            CONF.ml2_cisco_ucsm.ucsms['1.1.1.1'].sp_template_list, {})
        self.assertEqual(
            CONF.ml2_cisco_ucsm.ucsms['2.2.2.2'].sp_template_list,
            {"S1": ucsm_config.UCSTemplate("SP_Template1_path",
                                           "SP_Template1"),
             "S2": ucsm_config.UCSTemplate("SP_Template1_path",
                                           "SP_Template1"),
             "S3": ucsm_config.UCSTemplate("SP_Template2_path",
                                           "SP_Template2"),
             "S4": ucsm_config.UCSTemplate("SP_Template2_path",
                                           "SP_Template2")})

    def _assert_sp_templates_in_end_state(self):
        self.assertEqual(
            CONF.ml2_cisco_ucsm.ucsms['1.1.1.1'].sp_template_list,
            {"S1": ucsm_config.UCSTemplate("SP_Template1_path",
                                          "SP_Template1")})
        self.assertEqual(
            CONF.ml2_cisco_ucsm.ucsms['2.2.2.2'].sp_template_list,
            {"S2": ucsm_config.UCSTemplate("SP_Template1_path",
                                           "SP_Template1"),
             "S3": ucsm_config.UCSTemplate("SP_Template2_path",
                                           "SP_Template2"),
             "S4": ucsm_config.UCSTemplate("SP_Template2_path",
                                           "SP_Template2")})

    def test_add_sp_template_config_for_host(self):
        self._assert_sp_templates_in_start_state()
        ucsm_config.add_sp_template_config_for_host(
            'S1', '1.1.1.1', 'SP_Template1_path', 'SP_Template1')
        self._assert_sp_templates_in_end_state()

    def test_update_sp_template_config(self):
        self._assert_sp_templates_in_start_state()
        ucsm_config.update_sp_template_config(
            'S1', '1.1.1.1', 'SP_Template1_path/SP_Template1')
        self._assert_sp_templates_in_end_state()

    def test_support_pci_devices_bad_format(self):
        CONF.reset()
        nc_base.load_config_file(ucsm_config_bad_pci_devs)
        self.assertRaises(ValueError, CONF.ml2_cisco_ucsm.get,
                          "supported_pci_devs")


class ConfigMixin(object):

    """Mock config for UCSM driver."""

    mocked_parser = None

    def set_up_mocks(self):
        # Mock the configuration file
        CONF.reset()

        args = ['--config-file', base.etcdir('neutron.conf')]
        neutron_config.init(args=args)

        nc_base.load_config_file(ucsm_test_config_file)

        # Configure the ML2 mechanism drivers and network types
        ml2_opts = {
            'mechanism_drivers': ['cisco_ucsm'],
            'tenant_network_types': ['vlan'],
        }
        for opt, val in ml2_opts.items():
            ml2_config.cfg.CONF.set_override(opt, val, 'ml2')

        # Configure and test the Cisco UCS Manager mechanism driver
        ucsm_config.load_single_ucsm_config()



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

#######################################################
# Module that implements REST APIs for
# Cisco's CPNR (Cisco Prime Network Registrar) server
#######################################################

from neutron_lib import exceptions
from oslo_log import log as logging
from oslo_serialization import jsonutils
import requests
from requests import exceptions as r_exc

from networking_cisco._i18n import _LE

LOG = logging.getLogger(__name__)
URL_BASE = "web-services/rest/resource"


class CpnrException(exceptions.NeutronException):
    message = _LE("CPNR exception occurred")


class ServerError(CpnrException):
    message = _LE("CPNR received error response: %(status)i %(msg)s")


class Timeout(CpnrException):
    message = _LE("CPNR callout to server timed out: %(msg)s")


class ConnectionError(CpnrException):
    message = _LE("CPNR failed to connect: %(msg)s")


class UnexpectedError(CpnrException):
    message = _LE("CPNR unexpected error: %(msg)s")


class CpnrClient(object):
    """Class implementing REST APIs for CPNR Server."""

    def __init__(self, scheme, address, port, username,
                 password, insecure, timeout=20):
        self.url = scheme + "://" + address + ":" + str(port)
        self.auth = requests.auth.HTTPBasicAuth(username, password)
        self.headers = {'Content-Type': 'application/json',
                        'Accept': 'application/json'}
        self.dhcp_reload_needed = False
        self.dns_reload_needed = False
        self.verify_ssl_certs = not insecure
        self.timeout = timeout

    def __repr__(self):
        return "<CpnrClient(url=%s)>" % (self.url,)

    def _build_url(self, tags, vpn='', view='', zone=''):
        url = [self.url, URL_BASE]
        url.extend(tags)
        url = "/".join(url)
        if vpn:
            url += '?vpnId=' + vpn
        if view:
            url += '?viewId=' + view
        if zone:
            url += '&zoneOrigin=' + zone
        return url

    def _do_request(self, method, url, data=""):
        LOG.debug("%s %s %s", method, url, str(data))
        try:
            # Send request, receive response
            data = jsonutils.dumps(data) if data else None
            response = requests.request(method, url,
                                        data=data,
                                        auth=self.auth,
                                        headers=self.headers,
                                        timeout=self.timeout,
                                        verify=self.verify_ssl_certs)

            # Check for HTTP errors, raise exception
            response.raise_for_status()

            # Check if CPNR reload is needed
            if 'AX_DHCP_RELOAD_REQUIRED' in response.content:
                self.dhcp_reload_needed = True

            # If incomplete response, fetch next set of data
            if 'next' in response.links:
                nexturl = response.links['next']['url']
                nextresp = self._do_request(method, nexturl)
                return response.json() + nextresp

            return response.json() if method == 'GET' else None

        except r_exc.Timeout as te:
            raise Timeout(msg=te.message)
        except r_exc.ConnectionError as ce:
            raise ConnectionError(msg=ce.message)
        except r_exc.RequestException as re:
            raise UnexpectedError(msg=re.message)

    def get_dhcp_server(self):
        """Returns a dictionary with all the objects of DHCP server."""
        request_url = self._build_url(['DHCPServer'])
        return self._do_request('GET', request_url)

    def get_client_classes(self):
        """Returns a list of all the client classes from CPNR server."""
        request_url = self._build_url(['ClientClass'])
        return self._do_request('GET', request_url)

    def get_client_class(self, client_class_name):
        """Returns a specific client class details from CPNR server."""
        request_url = self._build_url(['ClientClass', client_class_name])
        return self._do_request('GET', request_url)

    def get_vpns(self):
        """Returns a list of all the VPNs from CPNR server."""
        request_url = self._build_url(['VPN'])
        return self._do_request('GET', request_url)

    def get_vpn(self, vpn_name):
        """Returns a specific VPN name details from CPNR server."""
        request_url = self._build_url(['VPN', vpn_name])
        return self._do_request('GET', request_url)

    def get_scopes(self, vpnid='.*'):
        """Returns a list of all the scopes from CPNR server."""
        request_url = self._build_url(['Scope'], vpn=vpnid)
        return self._do_request('GET', request_url)

    def get_scope(self, scope_name):
        """Returns a specific scope name details from CPNR server."""
        request_url = self._build_url(['Scope', scope_name])
        return self._do_request('GET', request_url)

    def get_client_entries(self):
        """Returns a list of all the client entries from CPNR server."""
        request_url = self._build_url(['ClientEntry'])
        return self._do_request('GET', request_url)

    def get_client_entry(self, client_entry_name):
        """Returns a specific client entry name details from CPNR server."""
        request_url = self._build_url(['ClientEntry', client_entry_name])
        return self._do_request('GET', request_url)

    def get_leases(self, vpnid='.*'):
        request_url = self._build_url(['Lease'], vpn=vpnid)
        return self._do_request('GET', request_url)

    def get_dns_server(self):
        request_url = self._build_url(['DNSServer'])
        return self._do_request('GET', request_url)

    def get_dns_forwarders(self):
        request_url = self._build_url(['DnsForwarder'])
        return self._do_request('GET', request_url)

    def get_dns_forwarder(self, name):
        request_url = self._build_url(['DnsForwarder', name])
        return self._do_request('GET', request_url)

    def get_dns_views(self):
        request_url = self._build_url(['DnsView'])
        return self._do_request('GET', request_url)

    def get_dns_view(self, name):
        request_url = self._build_url(['DnsView', name])
        return self._do_request('GET', request_url)

    def get_ccm_zones(self, viewid='.*'):
        request_url = self._build_url(['CCMZone'], view=viewid)
        return self._do_request('GET', request_url)

    def get_ccm_zone(self, name, viewid='.*'):
        request_url = self._build_url(['CCMZone', name], view=viewid)
        return self._do_request('GET', request_url)

    def get_ccm_reverse_zones(self, viewid='.*'):
        request_url = self._build_url(['CCMReverseZone'], view=viewid)
        return self._do_request('GET', request_url)

    def get_ccm_reverse_zone(self, name, viewid='.*'):
        request_url = self._build_url(['CCMReverseZone', name], view=viewid)
        return self._do_request('GET', request_url)

    def get_ccm_hosts(self, viewid='.*', zoneid='.*'):
        request_url = self._build_url(['CCMHost'], view=viewid, zone=zoneid)
        return self._do_request('GET', request_url)

    def get_ccm_host(self, name, viewid='.*', zoneid='.*'):
        request_url = self._build_url(['CCMHost', name],
                                      view=viewid, zone=zoneid)
        return self._do_request('GET', request_url)

    def create_scope(self, data):
        request_url = self._build_url(['Scope'])
        return self._do_request('POST', request_url, data)

    def create_client_class(self, data):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['ClientClass'])
        return self._do_request('POST', request_url, data)

    def create_vpn(self, data):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['VPN'])
        return self._do_request('POST', request_url, data)

    def create_client_entry(self, data):
        request_url = self._build_url(['ClientEntry'])
        return self._do_request('POST', request_url, data)

    def create_dns_forwarder(self, data):
        self.dns_reload_needed = True
        request_url = self._build_url(['DnsForwarder'])
        return self._do_request('POST', request_url, data)

    def create_dns_view(self, data):
        self.dns_reload_needed = True
        request_url = self._build_url(['DnsView'])
        return self._do_request('POST', request_url, data)

    def create_ccm_zone(self, data, viewid=None):
        self.dns_reload_needed = True
        request_url = self._build_url(['CCMZone'], view=viewid)
        return self._do_request('POST', request_url, data)

    def create_ccm_reverse_zone(self, data, viewid=None):
        self.dns_reload_needed = True
        request_url = self._build_url(['CCMReverseZone'], view=viewid)
        return self._do_request('POST', request_url, data)

    def create_ccm_host(self, data, viewid=None, zoneid=None):
        request_url = self._build_url(['CCMHost'], view=viewid, zone=zoneid)
        return self._do_request('POST', request_url, data)

    def update_dhcp_server(self, data):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['DHCPServer'])
        return self._do_request('PUT', request_url, data)

    def update_client_class(self, client_class_name, data):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['ClientClass', client_class_name])
        return self._do_request('PUT', request_url, data)

    def update_vpn(self, vpn_name, data):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['VPN', vpn_name])
        return self._do_request('PUT', request_url, data)

    def update_scope(self, scope_name, data):
        request_url = self._build_url(['Scope', scope_name])
        return self._do_request('PUT', request_url, data)

    def update_client_entry(self, client_entry_name, data):
        request_url = self._build_url(['ClientEntry', client_entry_name])
        return self._do_request('PUT', request_url, data)

    def update_dns_server(self, data):
        self.dns_reload_needed = True
        request_url = self._build_url(['DNSServer'])
        return self._do_request('PUT', request_url, data)

    def update_dns_forwarder(self, name, data):
        self.dns_reload_needed = True
        request_url = self._build_url(['DnsForwarder', name])
        return self._do_request('PUT', request_url, data)

    def update_dns_view(self, name, data):
        self.dns_reload_needed = True
        request_url = self._build_url(['DnsView', name])
        return self._do_request('PUT', request_url, data)

    def update_ccm_zone(self, name, data, viewid=None):
        self.dns_reload_needed = True
        request_url = self._build_url(['CCMZone', name], view=viewid)
        return self._do_request('PUT', request_url, data)

    def update_ccm_reverse_zone(self, name, data, viewid=None):
        self.dns_reload_needed = True
        request_url = self._build_url(['CCMReverseZone', name], view=viewid)
        return self._do_request('PUT', request_url, data)

    def update_ccm_host(self, name, data, viewid=None, zoneid=None):
        request_url = self._build_url(['CCMHost', name],
                                      view=viewid, zone=zoneid)
        return self._do_request('PUT', request_url, data)

    def delete_client_class(self, client_class_name):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['ClientClass', client_class_name])
        return self._do_request('DELETE', request_url)

    def delete_vpn(self, vpn_name):
        self.dhcp_reload_needed = True
        request_url = self._build_url(['VPN', vpn_name])
        return self._do_request('DELETE', request_url)

    def delete_scope(self, scope_name):
        request_url = self._build_url(['Scope', scope_name])
        return self._do_request('DELETE', request_url)

    def delete_client_entry(self, client_entry_name):
        request_url = self._build_url(['ClientEntry', client_entry_name])
        return self._do_request('DELETE', request_url)

    def delete_dns_forwarder(self, name):
        self.dns_reload_needed = True
        request_url = self._build_url(['DnsForwarder', name])
        return self._do_request('DELETE', request_url)

    def delete_dns_view(self, name):
        self.dns_reload_needed = True
        request_url = self._build_url(['DnsView', name])
        return self._do_request('DELETE', request_url)

    def delete_ccm_zone(self, name, viewid=None):
        request_url = self._build_url(['CCMZone', name], view=viewid)
        return self._do_request('DELETE', request_url)

    def delete_ccm_reverse_zone(self, name, viewid=None):
        self.dns_reload_needed = True
        request_url = self._build_url(['CCMReverseZone', name], view=viewid)
        return self._do_request('DELETE', request_url)

    def delete_ccm_host(self, name, viewid=None, zoneid=None):
        request_url = self._build_url(['CCMHost', name],
                                      view=viewid, zone=zoneid)
        return self._do_request('DELETE', request_url)

    def release_address(self, address, vpnid):
        """Release a specific lease, called after delete_client_entry"""
        query = address + "?action=releaseAddress&vpnId=" + vpnid
        request_url = self._build_url(['Lease', query])
        return self._do_request('DELETE', request_url)

    def reload_dhcp_server(self):
        try:
            request_url = self._build_url(['DHCPServer',
                                           '?action=reloadServer'])
            response = requests.request('PUT', request_url, auth=self.auth,
                                        verify=self.verify_ssl_certs)
            response.raise_for_status()
            self.dhcp_reload_needed = False
        except r_exc.RequestException as e:
            raise UnexpectedError(msg=e.message)

    def reload_dns_server(self):
        try:
            request_url = self._build_url(['DNSServer',
                                           '?action=reloadServer'])
            response = requests.request('PUT', request_url, auth=self.auth,
                                        verify=self.verify_ssl_certs)
            response.raise_for_status()
            self.dns_reload_needed = False
        except r_exc.RequestException as e:
            raise UnexpectedError(msg=e.message)

    def reload_needed(self):
        return self.dhcp_reload_needed or self.dns_reload_needed

    def reload_server(self, force_reload=False):
        if self.dhcp_reload_needed or force_reload:
            self.reload_dhcp_server()
        if self.dns_reload_needed or force_reload:
            self.reload_dns_server()

    def get_version(self):
        try:
            request_url = "/".join([self.url, "web-services/rest/session"])
            response = requests.request('GET', request_url,
                                        auth=self.auth, timeout=self.timeout,
                                        verify=self.verify_ssl_certs)
            return str(response.text)
        except r_exc.RequestException as e:
            raise UnexpectedError(msg=e.message)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
IOS-XE XML-based configuration snippets
"""

# The standard Template used to interact with IOS-XE.
# This template is added by the netconf client
# EXEC_CONF_SNIPPET = """
#       <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
#         <configure>
#           <__XML__MODE__exec_configure>%s
#           </__XML__MODE__exec_configure>
#         </configure>
#       </config>
# """

# =================================================
# Set ip address on an interface
# $(config)interface GigabitEthernet 1
# $(config)ip address 10.0.100.1 255.255.255.0
# =================================================
SET_INTC = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Enable an interface
# $(config)interface GigabitEthernet 1
# $(config)no shutdown
# =================================================
ENABLE_INTF = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no shutdown</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Disable an interface
# $(config)interface GigabitEthernet 1
# $(config)shutdown
# =================================================
DISABLE_INTF = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>shutdown</cmd>
        </cli-config-data>
</config>
"""
# =================================================
# Create VRF
# $(config)vrf definition nrouter-e7d4y5
# $(config-vrf)address-family ipv4
# $(config-vrf-af)exit-address-family
# $(config-vrf)address-family ipv6
# $(config-vrf-af)exit-address-family
# =================================================
CREATE_VRF = """
<config>
        <cli-config-data>
            <cmd>vrf definition %s</cmd>
            <cmd>address-family ipv4</cmd>
            <cmd>exit-address-family</cmd>
            <cmd>address-family ipv6</cmd>
            <cmd>exit-address-family</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Remove VRF
# $(config)no vrf definition nrouter-e7d4y5
# =================================================
REMOVE_VRF = """
<config>
        <cli-config-data>
            <cmd>no vrf definition %s</cmd>
        </cli-config-data>
</config>
"""
# =================================================
# Configure interface
# $(config)interface GigabitEthernet 2
# $(config)vrf forwarding nrouter-e7d4y5
# $(config)ip address 192.168.0.1 255.255.255.0
# =================================================
CONFIGURE_INTERFACE = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>ip address %s %s</cmd>
            <cmd>no shutdown</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Deconfigure Interface
# $(config)no interface GigabitEthernet 2
# =================================================
DECONFIGURE_INTERFACE = """
<config>
        <cli-config-data>
            <cmd>default interface %s</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Create Subinterface
# $(config)interface GigabitEthernet 2.500
# $(config)encapsulation dot1Q 500
# $(config)vrf forwarding nrouter-e7d4y5
# $(config)ip address 192.168.0.1 255.255.255.0
# =================================================
CREATE_SUBINTERFACE = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>

"""

# =================================================
# Remove Subinterface
# $(config)no interface GigabitEthernet 2.500
# =================================================
REMOVE_SUBINTERFACE = """
<config>
        <cli-config-data>
            <cmd>no interface %s</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Enable HSRP on a Subinterface
# $(config)interface GigabitEthernet 2.500
# $(config)vrf forwarding nrouter-e7d4y5
# $(config)standby version 2
# $(config)standby <group> priority <priority>
# $(config)standby <group> ip <ip>
# =================================================
SET_INTC_HSRP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>standby version 2</cmd>
            <cmd>standby %s priority %s</cmd>
            <cmd>standby %s ip %s</cmd>
        </cli-config-data>
</config>

"""

# =================================================
# Remove HSRP on a Subinterface
# $(config)interface GigabitEthernet 2.500
# $(config)no standby version 2
# $(config)no standby <group>
# =================================================
REMOVE_INTC_HSRP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no standby %s</cmd>
            <cmd>no standby version 2</cmd>
        </cli-config-data>
</config>

"""

# =================================================
# Create Access Control List
# $(config)ip access-list standard acl_500
# $(config)permit 192.168.0.1 255.255.255.0
# =================================================
CREATE_ACL = """
<config>
        <cli-config-data>
            <cmd>ip access-list standard %s</cmd>
            <cmd>permit %s %s</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Remove Access Control List
# $(config)no ip access-list standard acl_500
# =================================================
REMOVE_ACL = """
<config>
        <cli-config-data>
            <cmd>no ip access-list standard %s</cmd>
        </cli-config-data>
</config>
"""

# =========================================================================
# Set Dynamic source translation on an interface
# Syntax: ip nat inside source list <acl_no> interface <interface>
# .......vrf <vrf_name> overload
# eg: $(config)ip nat inside source list acl_500
#    ..........interface GigabitEthernet3.100 vrf nrouter-e7d4y5 overload
# =========================================================================
SNAT_CFG = "ip nat inside source list %s interface %s vrf %s overload"

SET_DYN_SRC_TRL_INTFC = """
<config>
        <cli-config-data>
            <cmd>ip nat inside source list %s interface %s vrf %s
            overload</cmd>
        </cli-config-data>
</config>

"""

# =========================================================================
# Remove Dynamic source translation on an interface
# Syntax: no ip nat inside source list <acl_no> interface <interface>
# .......vrf <vrf_name> overload
# eg: $(config)no ip nat inside source list acl_500
#    ..........interface GigabitEthernet3.100 vrf nrouter-e7d4y5 overload
# =========================================================================
REMOVE_DYN_SRC_TRL_INTFC = """
<config>
        <cli-config-data>
            <cmd>no ip nat inside source list %s interface %s vrf %s
            overload</cmd>
        </cli-config-data>
</config>

"""

# =================================================
# Set NAT
# Syntax : interface <interface>
#          ip nat <inside|outside>
# =================================================
SET_NAT = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>ip nat %s</cmd>
        </cli-config-data>
</config>
"""

# =================================================
# Remove NAT
# Syntax : interface <interface>
#          no ip nat <inside|outside>
# =================================================
REMOVE_NAT = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no ip nat %s</cmd>
        </cli-config-data>
</config>
"""

# =========================================================================
# Set Static source translation on an interface
# Syntax: ip nat inside source static <fixed_ip> <floating_ip>
# .......vrf <vrf_name> match-in-vrf
# eg: $(config)ip nat inside source static 192.168.0.1 121.158.0.5
#    ..........vrf nrouter-e7d4y5 match-in-vrf
# =========================================================================
SET_STATIC_SRC_TRL = """
<config>
        <cli-config-data>
            <cmd>ip nat inside source static %s %s vrf %s match-in-vrf</cmd>
        </cli-config-data>
</config>

"""

# =========================================================================
# Remove Static source translation on an interface
# Syntax: no ip nat inside source static <fixed_ip> <floating_ip>
# .......vrf <vrf_name> match-in-vrf
# eg: $(config)no ip nat inside source static 192.168.0.1 121.158.0.5
#    ..........vrf nrouter-e7d4y5 match-in-vrf
# =========================================================================
REMOVE_STATIC_SRC_TRL = """
<config>
        <cli-config-data>
            <cmd>no ip nat inside source static %s %s vrf %s match-in-vrf</cmd>
        </cli-config-data>
</config>

"""

# =============================================================================
# Set ip route
# Syntax: ip route vrf <vrf-name> <destination> <mask> [<interface>] <next hop>
# eg: $(config)ip route vrf nrouter-e7d4y5 8.8.0.0  255.255.0.0 10.0.100.255
# =============================================================================
SET_IP_ROUTE = """
<config>
        <cli-config-data>
            <cmd>ip route vrf %s %s %s %s</cmd>
        </cli-config-data>
</config>
"""

# =============================================================================
# Remove ip route
# Syntax: no ip route vrf <vrf-name> <destination> <mask>
#        [<interface>] <next hop>
# eg: $(config)no ip route vrf nrouter-e7d4y5 8.8.0.0  255.255.0.0 10.0.100.255
# =============================================================================
REMOVE_IP_ROUTE = """
<config>
        <cli-config-data>
            <cmd>no ip route vrf %s %s %s %s</cmd>
        </cli-config-data>
</config>
"""
# ===========================================================================
# Set default ip route
# Syntax: ip route vrf <vrf-name> 0.0.0.0 0.0.0.0 [<interface>] <next hop>
# eg: $(config)ip route vrf nrouter-e7d4y5 0.0.0.0  0.0.0.0 10.0.100.255
# ===========================================================================
DEFAULT_ROUTE_CFG = 'ip route vrf %s 0.0.0.0 0.0.0.0 %s'

SET_DEFAULT_ROUTE = """
<config>
        <cli-config-data>
            <cmd>ip route vrf %s 0.0.0.0 0.0.0.0 %s</cmd>
        </cli-config-data>
</config>
"""

# ===========================================================================
# Remove default ip route
# Syntax: ip route vrf <vrf-name> 0.0.0.0 0.0.0.0 [<interface>] <next hop>
# eg: $(config)ip route vrf nrouter-e7d4y5 0.0.0.0  0.0.0.0 10.0.100.255
# ===========================================================================
REMOVE_DEFAULT_ROUTE = """
<config>
        <cli-config-data>
            <cmd>no ip route vrf %s 0.0.0.0 0.0.0.0 %s</cmd>
        </cli-config-data>
</config>
"""

# =============================================================================
# Clear dynamic nat translations. This is used to clear any nat bindings before
# we can turn off NAT on an interface
# Syntax: clear ip nat translation [forced]
# =============================================================================
# CLEAR_DYN_NAT_TRANS = """
# <oper-data-format-text-block>
#     <exec>clear ip nat translation forced</exec>
# </oper-data-format-text-block>
# """
CLEAR_DYN_NAT_TRANS = """
<config>
        <cli-config-data>
            <cmd>do clear ip nat translation forced</cmd>
        </cli-config-data>
</config>
"""

## Additional utility snippets

#=========================================================================#
# Create nat pool
# Syntax: ip nat pool <pool_name> <source_ip> <dest_ip> prefix-length <30>
# eg: $(config)ip nat poolName 1.2.3.4 1.2.3.4 prefix 30
#========================================================================#
SET_NAT_POOL = """
<config>
        <cli-config-data>
            <cmd>ip nat pool %s %s %s prefix-length %s</cmd>
        </cli-config-data>
</config>

"""

#=========================================================================#
# Set Dynamic source translation on a nat pool
# Syntax: ip nat inside source list <acl_no> pool <pool_name>
# .......vrf <vrf_name> overload
# eg: $(config)ip nat inside source list acl_500
#    ..........pool pool_nrouter-e7d4y5 vrf nrouter-e7d4y5 overload
#========================================================================#
SET_DYN_SRC_TRL_POOL = """
<config>
        <cli-config-data>
            <cmd>ip nat inside source list %s pool %s vrf %s overload</cmd>
        </cli-config-data>
</config>

"""

#=========================================================================#
# Remove Dynamic source translation on a nat pool
# Syntax: no ip nat inside source list <acl_no> pool <pool_name>
# .......vrf <vrf_name> overload
# eg: $(config)no ip nat inside source list acl_500
#    ..........pool pool_nrouter-e7d4y5 vrf nrouter-e7d4y5 overload
#========================================================================#
REMOVE_DYN_SRC_TRL_POOL = """
<config>
        <cli-config-data>
            <cmd>no ip nat inside source list %s pool %s vrf %s overload</cmd>
        </cli-config-data>
</config>

"""



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class PluginSidePluggingDriver(object):
    """This class defines the API for plugging drivers.

    These are used used by Cisco (routing service) plugin to perform
    various operations on the logical ports of logical (service) resources
    in a plugin compatible way.
    """

    @abc.abstractmethod
    def create_hosting_device_resources(self, context, complementary_id,
                                        tenant_id, mgmt_context, max_hosted):
        """Create resources for a hosting device in a plugin specific way.

        Called when a hosting device is to be created so resources like
        networks and ports can be created for it in a plugin compatible
        way. This is primarily useful to service VMs.

        returns: a dict {'mgmt_port': <mgmt port or None>,
                         'ports': <list of ports>,
                         ... arbitrary driver items }

        :param context: Neutron api request context.
        :param complementary_id: complementary id of hosting device
        :param tenant_id: id of tenant owning the hosting device resources
        :param mgmt_context: dict with members:
             mgmt_ip_address: ip address of hosting device's management port
             mgmt_nw_id: id of management network for hosting devices
             mgmt_sec_grp_id: id of security group for management network
        :param max_hosted: maximum number of logical resources.
        """
        pass

    @abc.abstractmethod
    def get_hosting_device_resources(self, context, id, complementary_id,
                                     tenant_id, mgmt_nw_id):
        """Returns information about all resources for a hosting device.

        Called just before a hosting device is to be deleted so that
        information about the resources the hosting device uses can be
        collected.

        returns: a dict {'mgmt_port': <mgmt port or None>,
                         'ports': <list of ports>,
                         ... arbitrary driver items }

        :param context: Neutron api request context.
        :param id: id of hosting device.
        :param complementary_id: complementary id of hosting device
        :param tenant_id: id of tenant owning the hosting device resources.
        :param mgmt_nw_id: id of management network for hosting devices.
        """
        pass

    @abc.abstractmethod
    def delete_hosting_device_resources(self, context, tenant_id, mgmt_port,
                                        **kwargs):
        """Deletes resources for a hosting device in a plugin specific way.

        Called when a hosting device has been deleted (or when its creation
        has failed) so resources like networks and ports can be deleted in
        a plugin compatible way. This it primarily useful to service VMs.

        :param context: Neutron api request context.
        :param tenant_id: id of tenant owning the hosting device resources.
        :param mgmt_port: id of management port for the hosting device.
        :param kwargs: dictionary for any driver specific parameters.
        """
        pass

    @abc.abstractmethod
    def setup_logical_port_connectivity(self, context, port_db,
                                        hosting_device_id):
        """Establishes connectivity for a logical port.

        Performs the configuration tasks needed in the infrastructure
        to establish connectivity for a logical port.

        :param context: Neutron api request context.
        :param port_db: Neutron port that has been created.
        :param hosting_device_id: ID of the device where this port is attached.
        """
        pass

    @abc.abstractmethod
    def teardown_logical_port_connectivity(self, context, port_db,
                                           hosting_device_id):
        """Removes connectivity for a logical port.

        Performs the configuration tasks needed in the infrastructure
        to disconnect a logical port.

        Example: Remove a VLAN that is trunked to a service VM.

        :param context: Neutron api request context.
        :param port_db: Neutron port about to be deleted.
        :param hosting_device_id: ID of the device where this port is attached.
        """
        pass

    @abc.abstractmethod
    def extend_hosting_port_info(self, context, port_db, hosting_device,
                                 hosting_info):
        """Extends hosting information for a logical port.

        Allows a driver to add driver specific information to the
        hosting information for a logical port.

        :param context: Neutron api request context.
        :param port_db: Neutron port that hosting information concerns.
        :param hosting_device: Device that hosts the port.
        :param hosting_info: dict with hosting port information to be extended.
        """
        pass

    @abc.abstractmethod
    def allocate_hosting_port(self, context, router_id, port_db, network_type,
                              hosting_device_id):
        """Allocates a hosting port for a logical port.

        Schedules a logical port to a hosting port. Note that the hosting port
        may be the logical port itself.

        returns: a dict {'allocated_port_id': <id of allocated port>,
                         'allocated_vlan': <allocated VLAN or None>} or
                 None if allocation failed

        :param context: Neutron api request context.
        :param router_id: id of Neutron router the logical port belongs to.
        :param port_db: Neutron logical router port.
        :param network_type: Type of network for logical router port
        :param hosting_device_id: id of hosting device
        """
        pass



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def format_for_options(name, value):
    name = name.strip()
    if type(value) is str:
        value = value.strip()
    LOG.debug('name = %s value %s', name, value)
    if name not in OPTIONS:
        LOG.warning("Unrecognized DHCP options: %s", name)
        return
    code, datatype = OPTIONS[name]
    try:
        value = _format_value(datatype, value)
    except Exception:
        LOG.warning("Failed to parse DHCP option: %s", name)
        return
    value = ':'.join(value[i:i + 2] for i in range(0, len(value), 2))
    LOG.debug('name = %s value %s', name, value)
    return value


def _format_value(datatype, value):
    datatype = datatype.strip()
    if ',' in datatype:
        t1, _, t2 = datatype.partition(',')
        v1, _, v2 = value.partition(' ')
        return _format_value(t1, v1) + _format_value(t2, v2)
    elif datatype.endswith('-list'):
        t = datatype[:-5]
        return ''.join([_format_value(t, v) for v in value.split(',')])
    elif datatype == 'none':
        return ''
    elif datatype == 'bool':
        if value in set([True, 'True', 'yes', 'on', '1']):
            return '01'
        else:
            return '00'
    elif datatype.startswith('int'):
        length = int(datatype[3:]) / 4
        return ('{:0{}x}'.format(int(value), int(length)))
    elif datatype == 'string':
        return (binascii.hexlify(value.encode('utf-8'))).decode('utf-8')
    elif datatype == 'ip':
        return ''.join(['{:02x}'.format(int(o)) for o in value.split('.')])
    elif datatype == 'route':
        dest, _, nexthop = value.partition(' ')
        return _format_value('ip', dest) + _format_value('ip', nexthop)


def format_for_pnr(name, value):
    name = name.strip()
    value = value.strip()
    if name not in OPTIONS:
        LOG.warning("Unrecognized DHCP options: %s", name)
        return None
    code, datatype = OPTIONS[name]
    return {'number': str(code), 'value': value}


OPTIONS = {
    'subnet-mask': (1, 'ip'),
    'time-offset': (2, 'int32'),
    'routers': (3, 'ip-list'),
    'time-servers': (4, 'ip-list'),
    'name-servers': (5, 'ip-list'),
    'domain-name-servers': (6, 'ip-list'),
    'log-servers': (7, 'ip-list'),
    'quote-servers': (8, 'ip-list'),
    'lpr-servers': (9, 'ip-list'),
    'impress-servers': (10, 'ip-list'),
    'resource-location-servers': (11, 'ip-list'),
    'host-name': (12, 'string'),
    'boot-size': (13, 'int16'),
    'merit-dump': (14, 'string'),
    'domain-name': (15, 'string'),
    'swap-server': (16, 'ip'),
    'root-path': (17, 'string'),
    'extension-path': (18, 'string'),
    'ip-forwarding': (19, 'bool'),
    'non-local-source-routing': (20, 'bool'),
    'policy-filter': (21, 'ip-list'),
    'max-dgram-reassembly': (22, 'int16'),
    'default-ip-ttl': (23, 'int8'),
    'path-mtu-aging-timeout': (24, 'int32'),
    'path-mtu-plateau-table': (25, 'int16-list'),
    'interface-mtu': (26, 'int16'),
    'all-subnets-local': (27, 'bool'),
    'broadcast-address': (28, 'ip'),
    'perform-mask-discovery': (29, 'bool'),
    'mask-supplier': (30, 'bool'),
    'router-discovery': (31, 'bool'),
    'router-solicitation-address': (32, 'ip'),
    'static-routes': (33, 'ip-list'),
    'trailer-encapsulation': (34, 'bool'),
    'arp-cache-timeout': (35, 'int32'),
    'ieee802-3-encapsulation': (36, 'bool'),
    'default-tcp-ttl': (37, 'int8'),
    'tcp-keepalive-interval': (38, 'int32'),
    'tcp-keepalive-garbage': (39, 'bool'),
    'nis-domain': (40, 'string'),
    'nis-servers': (41, 'ip-list'),
    'ntp-servers': (42, 'ip-list'),
    'vendor-encapsulated-options': (43, 'string'),
    'netbios-name-servers': (44, 'ip-list'),
    'netbios-dd-server': (45, 'ip-list'),
    'netbios-node-type': (46, 'int8'),
    'netbios-scope': (47, 'string'),
    'font-servers': (48, 'ip-list'),
    'x-display-manager': (49, 'ip-list'),
    'dhcp-requested-address': (50, 'ip'),
    'dhcp-lease-time': (51, 'int32'),
    'dhcp-option-overload': (52, 'int8'),
    'dhcp-message-type': (53, 'int8'),
    'dhcp-server-identifier': (54, 'ip'),
    'dhcp-parameter-request-list': (55, 'string'),
    'dhcp-message': (56, 'string'),
    'dhcp-max-message-size': (57, 'int16'),
    'dhcp-renewal-time': (58, 'int32'),
    'dhcp-rebinding-time': (59, 'int32'),
    'class-id': (60, 'string'),
    'dhcp-client-identifier': (61, 'string'),
    'nwip-domain': (62, 'string'),
    'nwip-suboptions': (63, 'string'),
    'nisplus-domain': (64, 'string'),
    'nisplus-servers': (65, 'ip-list'),
    'tftp-server-name': (66, 'string'),
    'bootfile-name': (67, 'string'),
    'mobile-ip-home-agent': (68, 'ip-list'),
    'smtp-server': (69, 'ip-list'),
    'pop-server': (70, 'ip-list'),
    'nntp-server': (71, 'ip-list'),
    'www-server': (72, 'ip-list'),
    'finger-server': (73, 'ip-list'),
    'irc-server': (74, 'ip-list'),
    'streettalk-server': (75, 'ip-list'),
    'streettalk-directory-assistance-server': (76, 'ip-list'),
    'user-class': (77, 'string'),
    'slp-directory-agent': (78, 'int8,ip-list'),
    'slp-service-scope': (79, 'int8,string'),
    'rapid-commit': (80, 'none'),
    'client-fqdn': (81, 'string'),
    'storage-ns': (83, 'string'),
    'nds-servers': (85, 'ip-list'),
    'nds-tree-name': (86, 'string'),
    'nds-context': (87, 'string'),
    'bcms-controller-names': (88, 'string'),
    'bcms-controller-address': (89, 'string'),
    'dhcp-auth': (90, 'string'),
    'dhcp-client-last-time': (91, 'int32'),
    'associated-ip': (92, 'ip-list'),
    'system-architecture': (93, 'int16'),
    'interface-id': (94, 'string'),
    'ldap-servers': (95, 'ip-list'),
    'machine-id': (97, 'string'),
    'user-auth': (98, 'string'),
    'geoconf-civic': (99, 'string'),
    'ieee-1003-1-tz': (100, 'string'),
    'ref-tz-db': (101, 'string'),
    'netinfo-server-address': (112, 'string'),
    'netinfo-server-tag': (113, 'ip-list'),
    'default-url': (114, 'string'),
    'auto-configure': (116, 'bool'),
    'name-search': (117, 'int16-list'),
    'subnet-selection': (118, 'ip'),
    'domain-search': (119, 'string-list'),
    'sip-servers': (120, 'string'),
    'classless-static-routes': (121, 'route-list'),
    'dhcp-ccc': (122, 'string'),
    'dhcp-geoconf': (123, 'string'),
    'vendor-class-identifier': (124, 'string'),
    'vivso': (125, 'string'),
    'tftp-server': (128, 'ip-list'),
    'pxe-vendor-specific-129': (129, 'string'),
    'pxe-vendor-specific-130': (130, 'string'),
    'pxe-vendor-specific-131': (131, 'string'),
    'pxe-vendor-specific-132': (132, 'string'),
    'pxe-vendor-specific-133': (133, 'string'),
    'pxe-vendor-specific-134': (134, 'string'),
    'pxe-vendor-specific-135': (135, 'string'),
    'pana-agent': (136, 'ip-list'),
    'lost-server': (137, 'string'),
    'capwap-ac-v4': (138, 'ip-list'),
    'dhcp-mos': (139, 'string'),
    'dhcp-fqdn-mos': (140, 'string'),
    'sip-ua-config-domain': (141, 'string'),
    'andsf-servers': (142, 'ip-list'),
    'dhcp-geoloc': (144, 'string'),
    'force-renew-nonce-cap': (145, 'string'),
    'rdnss-selection': (146, 'string'),
    'tftp-server-address': (150, 'ip-list'),
    'status-code': (151, 'int8,string'),
    'dhcp-base-time': (152, 'int32'),
    'dhcp-state-start-time': (153, 'int32'),
    'dhcp-query-start-time': (154, 'int32'),
    'dhcp-query-end-time': (155, 'int32'),
    'dhcp-state': (156, 'int8'),
    'data-source': (157, 'int8'),
    'pcp-server': (158, 'string'),
    'dhcp-pxe-magic': (208, 'int32'),
    'config-file': (209, 'string'),
    'path-prefix': (210, 'string'),
    'reboot-time': (211, 'int32'),
    'dhcp-6rd': (212, 'string'),
    'dhcp-access-domain': (213, 'string'),
    'subnet-allocation': (220, 'string'),
    'dhcp-vss': (221, 'int8,string')}



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
import six

from neutron.db import agentschedulers_db

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import agent_exceptions
from networking_cisco.backwards_compatibility import get_agent_db_obj
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.db.device_manager.hd_models import (
    HostingDevice)
from networking_cisco.plugins.cisco.extensions import ciscocfgagentscheduler

LOG = logging.getLogger(__name__)


COMPOSITE_AGENTS_SCHEDULER_OPTS = [
    cfg.IntOpt('cfg_agent_down_time', default=30,
               help=_('Seconds of no status update until a cfg agent '
                      'is considered down.')),
    cfg.StrOpt('configuration_agent_scheduler_driver',
               default='networking_cisco.plugins.cisco.device_manager.'
                       'scheduler.hosting_device_cfg_agent_scheduler.'
                       'HostingDeviceCfgAgentScheduler',
               help=_('Driver to use for scheduling hosting device to a Cisco '
                      'configuration agent')),
    cfg.IntOpt('cfg_agent_monitoring_interval', default=20,
               help=("Maximal time (in seconds) between checks of config "
                     "agent status"))

]

cfg.CONF.register_opts(COMPOSITE_AGENTS_SCHEDULER_OPTS, "general")


class CfgAgentSchedulerDbMixin(
        ciscocfgagentscheduler.CfgAgentSchedulerPluginBase,
        agentschedulers_db.AgentSchedulerDbMixin):
    """Mixin class to add cfg agent scheduler extension."""

    # Scheduler of hosting devices to configuration agent
    cfg_agent_scheduler = None

    # Cache of statuses of cfg agents:
    # dict {<hd_id>: {'timestamp': <datetime.now()>
    _cfg_agent_statuses = {}

    @classmethod
    def is_agent_down(cls, heart_beat_time,
                      timeout=cfg.CONF.general.cfg_agent_down_time):
        return timeutils.is_older_than(heart_beat_time, timeout)

    def auto_schedule_hosting_devices(self, context, host):
        if self.cfg_agent_scheduler:
            cfg_agent_db = (
                self.cfg_agent_scheduler.auto_schedule_hosting_devices(
                    self, context, host))
            if not cfg_agent_db:
                return False
            with context.session.begin(subtransactions=True):
                query = context.session.query(HostingDevice)
                query = query.filter_by(cfg_agent_id=None)
                for hosting_device in query:
                    self._bind_hosting_device_to_cfg_agent(
                        context, hosting_device, cfg_agent_db)
            return True
        return False

    def assign_hosting_device_to_cfg_agent(self, context, cfg_agent_id,
                                           hosting_device_id):
        """Make config agent handle an (unassigned) hosting device."""
        hd_db = self._get_hosting_device(context, hosting_device_id)
        if hd_db.cfg_agent_id:
            if hd_db.cfg_agent_id == cfg_agent_id:
                return
            LOG.debug('Hosting device %(hd_id)s has already been assigned to '
                      'Cisco cfg agent %(agent_id)s',
                      {'hd_id': hosting_device_id, 'agent_id': cfg_agent_id})
            raise ciscocfgagentscheduler.HostingDeviceAssignedToCfgAgent(
                hosting_device_id=hosting_device_id, agent_id=cfg_agent_id)
        cfg_agent_db = get_agent_db_obj(self._get_agent(context, cfg_agent_id))
        if (cfg_agent_db.agent_type != c_constants.AGENT_TYPE_CFG or
                cfg_agent_db.admin_state_up is not True):
            raise ciscocfgagentscheduler.InvalidCfgAgent(agent_id=cfg_agent_id)
        self._bind_hosting_device_to_cfg_agent(context, hd_db, cfg_agent_db)
        cfg_notifier = self.agent_notifiers.get(c_constants.AGENT_TYPE_CFG)
        if cfg_notifier:
            cfg_notifier.hosting_devices_assigned_to_cfg_agent(
                context, [hosting_device_id], cfg_agent_db.host)

    def unassign_hosting_device_from_cfg_agent(self, context, cfg_agent_id,
                                               hosting_device_id):
        """Make config agent handle an (unassigned) hosting device."""
        hd_db = self._get_hosting_device(context, hosting_device_id)
        if hd_db.cfg_agent_id is None and cfg_agent_id is None:
            return
        elif hd_db.cfg_agent_id != cfg_agent_id:
            LOG.debug('Hosting device %(hd_id)s is not assigned to Cisco '
                      'cfg agent %(agent_id)s',
                      {'hd_id': hosting_device_id,
                       'agent_id': cfg_agent_id})
            raise ciscocfgagentscheduler.HostingDeviceNotAssignedToCfgAgent(
                hosting_device_id=hosting_device_id, agent_id=cfg_agent_id)
        cfg_agent_db = get_agent_db_obj(self._get_agent(context, cfg_agent_id))
        cfg_notifier = self.agent_notifiers.get(c_constants.AGENT_TYPE_CFG)
        if cfg_notifier:
            cfg_notifier.hosting_devices_unassigned_from_cfg_agent(
                context, [hosting_device_id], cfg_agent_db.host)
        self._bind_hosting_device_to_cfg_agent(context, hd_db, None)

    def list_hosting_devices_handled_by_cfg_agent(self, context, cfg_agent_id):
        return {'hosting_devices': self.get_hosting_devices(
            context, filters={'cfg_agent_id': [cfg_agent_id]})}

    def list_cfg_agents_handling_hosting_device(self, context,
                                                hosting_device_id):
        hd = self.get_hosting_device(context, hosting_device_id)
        cfg_agents = [self.get_agent(context, hd['cfg_agent_id'])] if hd[
            'cfg_agent_id'] else []
        return {'agents': cfg_agents}

    def get_cfg_agents(self, context, active=None, filters=None):
        query = context.session.query(bc.Agent)
        query = query.filter(
            bc.Agent.agent_type == c_constants.AGENT_TYPE_CFG)
        if active is not None:
            query = (query.filter(bc.Agent.admin_state_up == active))
        if filters:
            for key, value in six.iteritems(filters):
                column = getattr(bc.Agent, key, None)
                if column:
                    query = query.filter(column.in_(value))
        cfg_agents_db = query.all()
        if active is not None:
            cfg_agents_db = [
                cfg_agent_db for cfg_agent_db in cfg_agents_db
                if not self.is_agent_down(cfg_agent_db.heartbeat_timestamp)]
        return cfg_agents_db

    def get_cfg_agents_for_hosting_devices(self, context, hosting_device_ids,
                                           admin_state_up=None,
                                           schedule=False):
        if not hosting_device_ids:
            return []
        with context.session.begin(subtransactions=True):
            query = self.get_hosting_devices_qry(context, hosting_device_ids)
            if admin_state_up is not None:
                query = query.filter(
                    bc.Agent.admin_state_up == admin_state_up)
            agents = []
            agent_assigned_hd_ids = {}
            for hosting_device_db in query:
                current_agent_db = hosting_device_db.cfg_agent
                if current_agent_db:
                    self.set_monitor_timestamp(current_agent_db,
                                               timeutils.utcnow())
                    if (self.is_agent_down(
                            current_agent_db.heartbeat_timestamp) and
                            schedule):
                        # hosting device is handled by dead cfg agent so we'll
                        # try to reassign it to another cfg agent
                        LOG.info('Config agent %(agent_id)s is not alive. '
                                 'Un-assigning hosting device %(hd_id)s '
                                 'managed by it.',
                                 {'agent_id': current_agent_db.id,
                                  'hd_id': hosting_device_db.id})
                        hosting_device_db.cfg_agent = None
                if hosting_device_db.cfg_agent is None:
                    if schedule:
                        # only active cfg agents are considered by scheduler
                        agent_db = (
                            self.cfg_agent_scheduler.schedule_hosting_device(
                                self, context, hosting_device_db))
                        if agent_db is None:
                            continue
                        self._bind_hosting_device_to_cfg_agent(
                            context, hosting_device_db, agent_db)
                        agents.append(agent_db)
                        try:
                            agent_assigned_hd_ids[agent_db.id][
                                'hd_ids'].append(hosting_device_db.id)
                        except KeyError:
                            agent_assigned_hd_ids[agent_db.id] = {
                                'agent_host': agent_db.host,
                                'hd_ids': [hosting_device_db.id]}
                else:
                    agents.append(hosting_device_db.cfg_agent)
            self._notify_assignment(context, agent_assigned_hd_ids)
            return agents

    def _bind_hosting_device_to_cfg_agent(self, context, hosting_device_db,
                                          cfg_agent_db):
        with context.session.begin(subtransactions=True):
            if not hosting_device_db:
                LOG.debug('Hosting device to schedule not specified')
                return
            LOG.info('Assigning hosting device %(hd_id)s to config agent '
                     '%(agent_id)s.',
                     {'hd_id': hosting_device_db.id,
                      'agent_id':
                          cfg_agent_db.id if cfg_agent_db else None})
            hosting_device_db.cfg_agent = cfg_agent_db
            context.session.add(hosting_device_db)

    @lockutils.synchronized('devicemonitor', 'neutron-')
    def set_monitor_timestamp(self, agent, timestamp):
        LOG.debug('Setting monitoring time stamp for %s:' % agent['id'])
        self._cfg_agent_statuses[agent['id']] = {'timestamp': timestamp}

    def _sync_config_agent_monitoring(self, context):
        LOG.debug('Syncing monitored config agents')
        agents = self.get_agents(
            context, filters={'agent_type': [c_constants.AGENT_TYPE_CFG]},
            fields=['id'])
        self._cfg_agent_statuses = {agent['id']: {
            'timestamp': timeutils.utcnow()}
                                    for agent in agents}
        LOG.debug('Monitored config agents %s:' %
                  self._cfg_agent_statuses.keys())

    @lockutils.synchronized('devicemonitor', 'neutron-')
    def _check_config_agents(self):
        e_context = bc.context.get_admin_context()
        if not self._cfg_agent_statuses:
            self._sync_config_agent_monitoring(e_context)
        to_remove = []
        LOG.debug('In _check_config_agents: Monitored config agents %s:' %
                  self._cfg_agent_statuses.keys())
        for cfg_agent_id, info in six.iteritems(self._cfg_agent_statuses):
            if self.should_check_agent(info['timestamp']):
                # agent has not been checked recently so do it now
                LOG.debug('Must check status of config agent %s' %
                          cfg_agent_id)
                try:
                    agent = self.get_agent(e_context, cfg_agent_id)
                except agent_exceptions.AgentNotFound:
                    LOG.debug('Config agent %s does not exist anymore. Will '
                              'remove it from monitored config agents' %
                              cfg_agent_id)
                    to_remove.append(cfg_agent_id)
                    continue
                info['timestamp'] = timeutils.utcnow()
                if self.is_agent_down(agent['heartbeat_timestamp']):
                    LOG.info('Config agent %s is not alive. Un-assigning '
                             'hosting devices managed by it.',
                             cfg_agent_id)
                    self._reschedule_hosting_devices(e_context, cfg_agent_id)
        for cfg_agent_id in to_remove:
            LOG.debug('Removing config agent %s from monitored config agents' %
                      cfg_agent_id)
            del self._cfg_agent_statuses[cfg_agent_id]

    def _reschedule_hosting_devices(self, context, cfg_agent_id):
        with context.session.begin(subtransactions=True):
            agent_assigned_hd_ids = {}
            filters = {'cfg_agent_id': [cfg_agent_id]}
            for hd_db in self._get_collection_query(context, HostingDevice,
                                                    filters=filters):
                agent_db = self.cfg_agent_scheduler.schedule_hosting_device(
                    self, context, hd_db)
                LOG.info('Un-assigning hosting device %(hd_id)s from '
                         'config agent %(agent_id)s.',
                         {'hd_id': hd_db.id, 'agent_id': cfg_agent_id})
                if agent_db is None:
                    hd_db.cfg_agent_id = None
                else:
                    self._bind_hosting_device_to_cfg_agent(context, hd_db,
                                                           agent_db)
                    try:
                        agent_assigned_hd_ids[agent_db.id]['hd_ids'].append(
                            hd_db.id)
                    except KeyError:
                        agent_assigned_hd_ids[agent_db.id] = {
                            'agent_host': agent_db.host,
                            'hd_ids': [hd_db.id]}
        self._notify_assignment(context, agent_assigned_hd_ids)

    def _notify_assignment(self, context, agent_assigned_hd_ids):
        cfg_notifier = self.agent_notifiers.get(c_constants.AGENT_TYPE_CFG)
        if cfg_notifier:
            for agent_id, info in six.iteritems(agent_assigned_hd_ids):
                cfg_notifier.hosting_devices_assigned_to_cfg_agent(
                    context, info['hd_ids'], info['agent_host'])

    @classmethod
    def should_check_agent(
            cls, heart_beat_time,
            timeout=cfg.CONF.general.cfg_agent_monitoring_interval):
        return timeutils.is_older_than(heart_beat_time, timeout)



# Copyright 2017 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Topology Discovery constants:

# Query LLDP Daemon every 15 seconds.
PERIODIC_TASK_INTERVAL = 15
# This means a topology update message will be sent after every minute (15*4),
# even if there's no change in the parameters.
TOPO_DISC_SEND_THRESHOLD = 4



# Copyright (c) 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import mock
from oslo_config import cfg

from networking_cisco.backwards_compatibility import constants as p_const
from networking_cisco.backwards_compatibility import ml2_api as api

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import exceptions
from networking_cisco.ml2_drivers.nexus import nexus_db_v2

from networking_cisco.tests.unit.ml2_drivers.nexus import (
    test_cisco_nexus_base)


class TestCiscoNexusVxlanDeviceConfig(object):
    """Config Data for Cisco ML2 VXLAN Nexus device driver."""

    test_configs = {
        'test_vxlan_config1':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_1,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.VXLAN_ID,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_vxlan_config2':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                test_cisco_nexus_base.HOST_NAME_2,
                test_cisco_nexus_base.NEXUS_PORT_2,
                test_cisco_nexus_base.INSTANCE_2,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.VXLAN_ID,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_vxlan_config3':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_6,
                test_cisco_nexus_base.HOST_NAME_3,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.VXLAN_ID,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_vxlan_config4':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_7,
                test_cisco_nexus_base.HOST_NAME_3,
                test_cisco_nexus_base.NEXUS_DUAL_2,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.VXLAN_ID,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_vxlan_config5':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_8,
                test_cisco_nexus_base.HOST_NAME_4,
                test_cisco_nexus_base.NEXUS_PORT_1,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_1,
                test_cisco_nexus_base.VXLAN_ID,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
        'test_vxlan_config6':
            test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_8,
                test_cisco_nexus_base.HOST_NAME_5,
                test_cisco_nexus_base.NEXUS_PORT_2,
                test_cisco_nexus_base.INSTANCE_1,
                test_cisco_nexus_base.VLAN_ID_2,
                test_cisco_nexus_base.VXLAN_ID + 1,
                test_cisco_nexus_base.MCAST_GROUP,
                test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                {},
                None,
                test_cisco_nexus_base.NORMAL_VNIC),
    }

    test_configs = collections.OrderedDict(sorted(test_configs.items()))


class TestCiscoNexusVxlanResults(
    test_cisco_nexus_base.TestCiscoNexusBaseResults):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {

        # The following contains desired Nexus output for
        # some basic config above.
        'add_port_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_NVE_INTERFACE.
                format(1, 70000, '255.1.1.1'),
            test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(267, 70000),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)]),

        'delete_port_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_NVE_INTERFACE.
                format(1, 70000, 267),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)]),

        'add_port2_driver_result': (
            [test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(267, 70000),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'delete_port2_driver_result': (
            [test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/20', 267)]),

        'add_port_driver_result3': (
            [test_cisco_nexus_base.RESULT_ADD_NVE_INTERFACE.
                format(1, 70000, '255.1.1.1'),
            test_cisco_nexus_base.RESULT_ADD_NVE_INTERFACE.
                format(1, 70000, '255.1.1.1'),
            test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(267, 70000),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(267, 70000),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/2', 267),
            test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(267, 70000),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/3', 267)]),

        'delete_port_driver_result3': (
            [test_cisco_nexus_base.RESULT_DEL_NVE_INTERFACE.
                format(1, 70000, 267),
             test_cisco_nexus_base.RESULT_DEL_NVE_INTERFACE.
                format(1, 70000, 267),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/2', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/3', 267)]),

        'add_port_driver_result2': (
            [test_cisco_nexus_base.RESULT_ADD_NVE_INTERFACE.
                format(1, 70001, '255.1.1.1'),
            test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(265, 70001),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/20', 265)]),

        'delete_port_driver_result2': (
            [test_cisco_nexus_base.RESULT_DEL_NVE_INTERFACE.
                format(1, 70001, 265),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/20', 265),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(265)]),

        'add_port_driver_result4': (
            [test_cisco_nexus_base.RESULT_ADD_NVE_INTERFACE.
                format(1, 70000, '255.1.1.1'),
            test_cisco_nexus_base.RESULT_ADD_VLAN_VNI.
                format(267, 70000),
            test_cisco_nexus_base.RESULT_ADD_INTERFACE.
                format('ethernet', '1\/10', 267)]),

        'delete_port_driver_result4': (
            [test_cisco_nexus_base.RESULT_DEL_NVE_INTERFACE.
                format(1, 70000, 267),
            test_cisco_nexus_base.RESULT_DEL_INTERFACE.
                format('ethernet', '1\/10', 267),
            test_cisco_nexus_base.RESULT_DEL_VLAN.format(267)])

    }


class TestCiscoNexusVxlanDevice(test_cisco_nexus_base.TestCiscoNexusBase,
                           TestCiscoNexusVxlanDeviceConfig):

    """Unit tests for Cisco ML2 VXLAN Nexus device driver."""

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        cfg.CONF.set_override('nexus_driver', 'ncclient', 'ml2_cisco')
        cfg.CONF.set_override('never_cache_ssh_connection', False, 'ml2_cisco')
        cfg.CONF.set_override('switch_heartbeat_time', 0, 'ml2_cisco')
        super(TestCiscoNexusVxlanDevice, self).setUp()
        self.mock_ncclient.reset_mock()
        self.addCleanup(self._clear_nve_db)
        self.results = TestCiscoNexusVxlanResults()

    def _clear_nve_db(self):
        nexus_db_v2.remove_all_nexusnve_bindings()

    def test_enable_vxlan_feature_failure(self):
        """Verifies exception during enable VXLAN driver. """

        # Set configuration variable to add/delete the VXLAN global nexus
        # switch values.
        cfg.CONF.set_override('vxlan_global_config', True, 'ml2_cisco')
        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'feature nv overlay vn-segment-vlan-based',
            'test_vxlan_config1',
            __name__)

    def test_disable_vxlan_feature_failure(self):
        """Verifies exception during disable VXLAN driver. """

        # Set configuration variable to add/delete the VXLAN global nexus
        # switch values.
        cfg.CONF.set_override('vxlan_global_config', True, 'ml2_cisco')
        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'no feature nv overlay vn-segment-vlan-based',
            'test_vxlan_config1',
            __name__)

    def test_create_nve_member_failure(self):
        """Verifies exception during create nve member driver. """

        self._create_port_failure(
            'connect.return_value.edit_config.side_effect',
            'member vni mcast-group',
            'test_vxlan_config1',
            __name__)

    def test_delete_nve_member_failure(self):
        """Verifies exception during delete nve member driver. """

        self._delete_port_failure(
            'connect.return_value.edit_config.side_effect',
            'no member vni',
            'test_vxlan_config1',
            __name__)

    def test_nexus_vxlan_one_network_two_hosts(self):
        """Tests creation and deletion of two new virtual Ports."""

        self._basic_create_verify_port_vlan(
            'test_vxlan_config1',
            self.results.get_test_results(
                'add_port_driver_result'))

        self._create_port(
            self.test_configs['test_vxlan_config2'])
        self._verify_results(
            self.results.get_test_results(
                'add_port2_driver_result'))

        bindings = nexus_db_v2.get_nexusvlan_binding(
                       test_cisco_nexus_base.VLAN_ID_1,
                       test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)
        self.assertEqual(2, len(bindings))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_vxlan_config2',
            self.results.get_test_results(
                'delete_port2_driver_result'),
            nbr_of_bindings=1)

        self._basic_delete_verify_port_vlan(
            'test_vxlan_config1',
            self.results.get_test_results(
                'delete_port_driver_result'))

    def test_nexus_missing_vxlan_fields(self):
        """Test handling of a VXLAN NexusMissingRequiredFields exception.

        Test the Cisco NexusMissingRequiredFields exception by using
        empty VNI and mcast address values during port update event.
        """
        local_test_configs = {
            'test_vxlan_config_no_vni':
                test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                    test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                    test_cisco_nexus_base.HOST_NAME_1,
                    test_cisco_nexus_base.NEXUS_PORT_1,
                    test_cisco_nexus_base.INSTANCE_1,
                    test_cisco_nexus_base.VLAN_ID_1,
                    None,
                    test_cisco_nexus_base.MCAST_GROUP,
                    test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                    {},
                    None,
                    test_cisco_nexus_base.NORMAL_VNIC),
            'test_vxlan_config_no_mcast':
                test_cisco_nexus_base.TestCiscoNexusBase.TestConfigObj(
                    test_cisco_nexus_base.NEXUS_IP_ADDRESS_1,
                    test_cisco_nexus_base.HOST_NAME_1,
                    test_cisco_nexus_base.NEXUS_PORT_1,
                    test_cisco_nexus_base.INSTANCE_1,
                    test_cisco_nexus_base.VLAN_ID_1,
                    test_cisco_nexus_base.VXLAN_ID,
                    None,
                    test_cisco_nexus_base.DEVICE_OWNER_COMPUTE,
                    {},
                    None,
                    test_cisco_nexus_base.NORMAL_VNIC),
        }
        test_list = ('test_vxlan_config_no_vni',
                     'test_vxlan_config_no_mcast')
        for test_name in test_list:
            self.assertRaises(
                exceptions.NexusMissingRequiredFields,
                self._create_port,
                local_test_configs[test_name])

    def test_nexus_vxlan_bind_port(self):
        """Test VXLAN bind_port method processing.

        Verify the bind_port method allocates the VLAN segment correctly.
        """
        expected_dynamic_segment = {
            api.SEGMENTATION_ID: mock.ANY,
            api.PHYSICAL_NETWORK: test_cisco_nexus_base.PHYSNET,
            api.ID: mock.ANY,
            api.NETWORK_TYPE: p_const.TYPE_VLAN}

        self.mock_get_dynamic_segment.return_value = expected_dynamic_segment

        self._bind_port(self.test_configs['test_vxlan_config1'])
        self.mock_continue_binding.assert_called_once_with(
            test_cisco_nexus_base.NETID,
            [expected_dynamic_segment])

    def test_nexus_vxlan_bind_port_no_physnet(self):
        """Test VXLAN bind_port error processing.

        Verify that continue_binding() method is not called when no 'physnet'
        key is present in the nexus switch dictionary.
        """

        cfg.CONF.set_override(
            const.PHYSNET, None,
            cfg.CONF.ml2_cisco.nexus_switches.get(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_1)._group)

        try:
            self._bind_port(self.test_configs['test_vxlan_config1'])
        except exceptions.PhysnetNotConfigured:
            assert not self.mock_continue_binding.called

    def test_nexus_vxlan_bind_port_no_dynamic_segment(self):
        """Test VXLAN bind_port processing.

        Verify that the continue_binding() method is not called when the vlan
        dynamic segment wasn't allocated.
        """

        self.mock_get_dynamic_segment.return_value = None

        try:
            self._bind_port(self.test_configs['test_vxlan_config1'])
        except exceptions.NoDynamicSegmentAllocated:
            assert not self.mock_continue_binding.called

    def test_nexus_vxlan_one_network(self):
        """Test processing for creating one VXLAN segment."""

        # Since test_vxlan_config3 & test_vxlan_config4 share
        # the same host name they both get processed in the
        # next call.
        self._basic_create_verify_port_vlan(
            'test_vxlan_config3',
            self.results.get_test_results(
                'add_port_driver_result3'))

        for switch_ip, nbr_bind in [
            (test_cisco_nexus_base.NEXUS_IP_ADDRESS_6, 1),
            (test_cisco_nexus_base.NEXUS_IP_ADDRESS_7, 2)]:
            bindings = nexus_db_v2.get_nexusvlan_binding(
                           test_cisco_nexus_base.VLAN_ID_1,
                           switch_ip)
            self.assertEqual(nbr_bind, len(bindings))
            binding = nexus_db_v2.get_nve_switch_bindings(switch_ip)
            self.assertEqual(1, len(binding))

        # Since test_vxlan_config3 & test_vxlan_config4 share
        # the same host name they both get processed in the
        # next call.
        self._basic_delete_verify_port_vlan(
            'test_vxlan_config3',
            self.results.get_test_results(
                'delete_port_driver_result3'))

        for switch_ip in [
            test_cisco_nexus_base.NEXUS_IP_ADDRESS_6,
            test_cisco_nexus_base.NEXUS_IP_ADDRESS_7]:
            try:
                bindings = nexus_db_v2.get_nexusvlan_binding(
                               test_cisco_nexus_base.VLAN_ID_1,
                               switch_ip)
            except exceptions.NexusPortBindingNotFound:
                bindings = []
            self.assertEqual(0, len(bindings))
            try:
                binding = nexus_db_v2.get_nve_switch_bindings(switch_ip)
            except exceptions.NexusPortBindingNotFound:
                binding = []
            self.assertEqual(0, len(binding))

    def test_nexus_vxlan_two_network(self):
        """Test processing for creating one VXLAN segment."""

        self._basic_create_verify_port_vlan(
            'test_vxlan_config5',
            self.results.get_test_results(
                'add_port_driver_result4'))

        self._create_port(
            self.test_configs['test_vxlan_config6'],
            override_netid=888)
        self._verify_results(
            self.results.get_test_results(
                'add_port_driver_result2'))

        binding = nexus_db_v2.get_nve_switch_bindings(
            test_cisco_nexus_base.NEXUS_IP_ADDRESS_8)
        self.assertEqual(2, len(binding))

        # Clean all the ncclient mock_calls so we can evaluate
        # results of delete operations.
        self.mock_ncclient.reset_mock()

        self._basic_delete_verify_port_vlan(
            'test_vxlan_config6',
            self.results.get_test_results(
                'delete_port_driver_result2'),
            nbr_of_bindings=1)

        self._basic_delete_verify_port_vlan(
            'test_vxlan_config5',
            self.results.get_test_results(
                'delete_port_driver_result4'))

        try:
            binding = nexus_db_v2.get_nve_switch_bindings(
                test_cisco_nexus_base.NEXUS_IP_ADDRESS_8)
        except exceptions.NexusPortBindingNotFound:
            binding = []
        self.assertEqual(0, len(binding))
